﻿using System;
using System.ComponentModel;
using System.Runtime.InteropServices;
using glynor.Common.Utilities;
using MediaCenter;
using Timer = System.Timers.Timer;

namespace MCnet.COM
{
    internal static class MCComConnector
    {
        #region Private Members

        // NLog Class Logger
        private static readonly NLog.Logger ClassLogger = NLog.LogManager.GetCurrentClassLogger();

        // Internal Instance
        private static IMJAutomation _instance;
        // This is the padlock that enables thread safety
        private static readonly object InstanceLock = new object();

        private static readonly Timer ConnectionTimer = new Timer();

        // How long for the connector in GetMC() to wait before giving up (in milliseconds)
        private static int _connectionTimeout = 120000;

        #endregion

        #region Instance

        /// <summary>
        /// This is the outward-facing singleton instance of IMJAutomation.
        /// </summary>
        public static IMJAutomation Instance
        {
            get
            {
                // If it isn't null, just return it.
                if (_instance != null) return _instance;
                // Otherwise we might need to create it, so lock.
                lock (InstanceLock)
                {
                    if (_instance == null)
                    {
                        _instance = GetMC();
                    }
                    return _instance;
                }
            }
        }

        #endregion

        #region Properties

        /// <summary>
        /// Returns True if the Instance has already been created.
        /// </summary>
        public static Boolean InstanceExists
        {
            get { return _instance != null; }
        }

        /// <summary>
        /// Returns True if connecting to the instance required launching MC.
        /// </summary>
        public static Boolean LaunchedMC { get; private set; }

        /// <summary>
        /// The Connection Timeout in milliseconds. (Set to zero to disable.)
        /// </summary>
        /// <remarks>
        /// By default, MCComConnector waits 120 seconds before giving up on the connection to the Automation object.
        /// If you want to change this, you need to do it before calling the MCComConnector.Instance property.
        /// </remarks>
        [DefaultValue(120000)]
        public static int ConnectionTimeout
        {
            set
            {
                if (value < 0)
                {
                    throw new ArgumentOutOfRangeException("value", "The ConnectionTimeout cannot be a negative value.");
                }
                _connectionTimeout = value;
            }
            get { return _connectionTimeout; }
        }

        #endregion

        #region Startup Methods

        public static Boolean Connect()
        {
            // If we're already connected, just return
            if (InstanceExists) return true;

            // Then just try to use the external Instance and it'll connect.
            return Instance != null;
        }

        /// <summary>
        /// Gets the MC Core Automation object
        /// This is used ONLY by the Instance Property
        /// </summary>
        private static IMJAutomation GetMC()
        {
            // Run the connection timer
            RunConnectionTimer();
            // Get the MC Instance!
            MCAutomation mcAutomation;
            try
            {
                // Try to connect to a running instance of MC
                mcAutomation = GetRunningMC();
            }
            catch (Exception e)
            {
                // Check the type of exception we got.  Certain types can be handled.
                switch (e.HResult)
                {
                    case -2147221005:
                    {
                        throw new InvalidOperationException(
                            "MC isn't registered properly in COM.  Reinstall the current version of Media Center.", e);
                    }
                    case -2147221021:
                    {
                        // Then the exception was "Operation unavailable (Exception from HRESULT: 0x800401E3 (MK_E_UNAVAILABLE))"
                        // Which is what happens when MC isn't currently running.
                        ClassLogger.Info("No running instance of MC found.");
                        // Try to create an instance of MC instead
                        mcAutomation = LaunchMC();
                    }
                        break;
                    default:
                    {
                        // Then we don't know what the error was, so throw it.
                        throw;
                    }
                }
            }
            return mcAutomation;
        }

        private static void RunConnectionTimer()
        {
            // If the ConnectionTimeout setting is Zero, then the timeout is disabled.
            if (ConnectionTimeout == 0) return;

            // Otherwise, set a timer in case the connection is never completed
            // Defaults to 120 seconds.  You can change this using the MC.ConnectionTimeout property
            ConnectionTimer.Elapsed += ThrowConnectionTimedOut;
            ConnectionTimer.AutoReset = false;
            ConnectionTimer.Interval = ConnectionTimeout;
            // Run the timer
            ConnectionTimer.Enabled = true;
        }

        /// <summary>
        /// This attempts to connect to an already running copy of MC.
        /// </summary>
        /// <returns>The MCAutomation instance for the connected copy of MC.</returns>
        private static MCAutomation GetRunningMC()
        {
            ClassLogger.Info("Attempting to connect to running instance of MC.");
            // Try to connect to a running instance of MC
            var mcAutomation = (MCAutomation)Marshal.GetActiveObject("MediaJukebox Application");
            if (ValidateInstance(mcAutomation))
            {
                // Kill the connection timer
                ConnectionTimer.Enabled = false;
                ClassLogger.Info("Successfully connected to Media Center.");
                return mcAutomation;
            }
            else return null;  // This will never get returned because ValidateInstance will throw an exception.
        }

        /// <summary>
        /// This attempts to launch MC and returns the MCAutomation instance.
        /// </summary>
        /// <returns>The MCAutomation instance after launching.</returns>
        private static MCAutomation LaunchMC()
        {
            ClassLogger.Info("Launching Media Center...");
            try
            {
                var mcAutomation =
                    (MCAutomation) Activator.CreateInstance(Type.GetTypeFromProgID("MediaJukebox Application"));
                if (ValidateInstance(mcAutomation))
                {
                    // Kill the connection timer
                    ConnectionTimer.Enabled = false;
                    ClassLogger.Info("Successfully launched Media Center.");
                    LaunchedMC = true;
                    return mcAutomation;
                }
                else return null;  // This will never get returned because ValidateInstance will throw an exception.
            }
            catch (Exception ex)
            {
                // Something terrible happened
                ConnectionTimer.Enabled = false;
                ClassLogger.Fatal("Launching MC Failed: {0}", ex.Message);
                ClassLogger.Fatal(ex);
                throw;
            }
        }

        private static Boolean ValidateInstance(MCAutomation instance)
        {
            if (instance.IVersion >= 2)
            {
                return true;
            }
            else
            {
                // Wrong IVersion
                _instance = null;
                ConnectionTimer.Enabled = false;
                throw new InvalidOperationException(
                    string.Format(
                        "MC instance doesn't support the required Automation Interface (Required: 2, Received: {0})",
                        instance.IVersion));
            }
        }

        private static void ThrowConnectionTimedOut(object source, System.Timers.ElapsedEventArgs e)
        {
            if (_instance == null)
            {
                // The connection must have failed or not returned (probably due to Auto-Update Installer)
                throw new InvalidOperationException(string.Format("MC failed to load within timeout: {0} seconds.",
                    TimeSpanUtil.ConvertMillisecondsToSeconds(ConnectionTimeout)));
            }
        }

        #endregion
    }
}
